/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2012 Mudit Raj Gupta
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This class handles all the callbacks for the ns3Federate. On each callback, it will
 * just create a string describing the event and add that string to a list of events that have
 * happened. At the end of processing, the contents of this list can be printed so we can see
 * exactly what events happened in what order.
 *
 * Authors: Mudit Raj Gupta <mudit.raaj.gupta@gmail.com>
 *
 */

import hla.rti.ArrayIndexOutOfBounds;
import hla.rti.EventRetractionHandle;
import hla.rti.LogicalTime;
import hla.rti.ReceivedInteraction;
import hla.rti.ReflectedAttributes;
import hla.rti.jlc.EncodingHelpers;
import hla.rti.jlc.NullFederateAmbassador;

import org.portico.impl.hla13.types.DoubleTime;

/**
 * This class handles all incoming callbacks from the RTI regarding a particular
 * Federate. It will log information about any callbacks it
 * receives, thus demonstrating how to deal with the provided callback information.
 */
public class dummyFederateAmbassador extends NullFederateAmbassador
{
	// these variables are accessible in the package
	protected double federateTime        = 0.0;
	protected double federateLookahead   = 1.0;
	
        protected boolean isRegulating       = false;
	protected boolean isConstrained      = false;
	protected boolean isAdvancing        = false;
	
	protected boolean isAnnounced        = false;
	protected boolean isReadyToRun       = false;

	public dummyFederateAmbassador()
	{
	}

	
	private double convertTime( LogicalTime logicalTime )
	{
		// PORTICO SPECIFIC!!
		return ((DoubleTime)logicalTime).getTime();
	}
	
	private void log( String message )
	{
		System.out.println( "FederateAmbassador: " + message );
	}
	
	
	public void synchronizationPointRegistrationFailed( String label )
	{
		log( "Failed to register sync point: " + label );
	}

	public void synchronizationPointRegistrationSucceeded( String label )
	{
		log( "Successfully registered sync point: " + label );
	}

	public void announceSynchronizationPoint( String label, byte[] tag )
	{
		log( "Synchronization point announced: " + label );
		if( label.equals(dummyFederate.READY_TO_RUN) )
			this.isAnnounced = true;
	}

	public void federationSynchronized( String label )
	{
		log( "Federation Synchronized: " + label );
		if( label.equals(dummyFederate.READY_TO_RUN) )
			this.isReadyToRun = true;
	}

	/**
	 * The RTI has informed us that time regulation is now enabled.
	 */
	public void timeRegulationEnabled( LogicalTime theFederateTime )
	{
		this.federateTime = convertTime( theFederateTime );
		this.isRegulating = true;
	}

	public void timeConstrainedEnabled( LogicalTime theFederateTime )
	{
		this.federateTime = convertTime( theFederateTime );
		this.isConstrained = true;
	}

	public void timeAdvanceGrant( LogicalTime theTime )
	{
		this.federateTime = convertTime( theTime );
		this.isAdvancing = false;
	}

	public void discoverObjectInstance( int theObject,
	                                    int theObjectClass,
	                                    String objectName )
	{
		log( "Discoverd Object: handle=" + theObject + ", classHandle=" +
		     theObjectClass + ", name=" + objectName );
	}

	public void reflectAttributeValues( int theObject,
	                                    ReflectedAttributes theAttributes,
	                                    byte[] tag )
	{
		// just pass it on to the other method for printing purposes
		// passing null as the time will let the other method know it
		// it from us, not from the RTI
		reflectAttributeValues( theObject, theAttributes, tag, null, null );
	}

	public void reflectAttributeValues( int theObject,
	                                    ReflectedAttributes theAttributes,
	                                    byte[] tag,
	                                    LogicalTime theTime,
	                                    EventRetractionHandle retractionHandle )
	{
		StringBuilder builder = new StringBuilder( "Reflection for object:" );
		System.out.print("test");
		// print the handle
		builder.append( " handle=" + theObject );
		// print the tag
		builder.append( ", tag=" + EncodingHelpers.decodeString(tag) );
		// print the time (if we have it) we'll get null if we are just receiving
		// a forwarded call from the other reflect callback above
		if( theTime != null )
		{
			builder.append( ", time=" + convertTime(theTime) );
		}
		
		// print the attribute information
		builder.append( ", attributeCount=" + theAttributes.size() );
		builder.append( "\n" );
		for( int i = 0; i < theAttributes.size(); i++ )
		{
			try
			{
				// print the attibute handle
				builder.append( "\tattributeHandle=" );
				builder.append( theAttributes.getAttributeHandle(i) );
				// print the attribute value
				builder.append( ", attributeValue=" );
				builder.append(
				    EncodingHelpers.decodeString(theAttributes.getValue(i)) );
				builder.append( "\n" );
			}
			catch( ArrayIndexOutOfBounds aioob )
			{
				// won't happen
			}
		}
		
		log( builder.toString() );
	}

	public void receiveInteraction( int interactionClass,
	                                ReceivedInteraction theInteraction,
	                                byte[] tag )
	{
		// just pass it on to the other method for printing purposes
		// passing null as the time will let the other method know it
		// it from us, not from the RTI
		receiveInteraction( interactionClass, theInteraction, tag, null, null );
	}

	public void receiveInteraction( int interactionClass,
	                                ReceivedInteraction theInteraction,
	                                byte[] tag,
	                                LogicalTime theTime,
	                                EventRetractionHandle eventRetractionHandle )
	{
		StringBuilder builder = new StringBuilder( "Interaction Received:" );
		
		// print the handle
		builder.append( " handle=" + interactionClass );
		// print the tag
		builder.append( ", tag=" + EncodingHelpers.decodeString(tag) );
		// print the time (if we have it) we'll get null if we are just receiving
		// a forwarded call from the other reflect callback above
		if( theTime != null )
		{
			builder.append( ", time=" + convertTime(theTime) );
		}
		
		// print the parameer information
		builder.append( ", parameterCount=" + theInteraction.size() );
		builder.append( "\n" );
		for( int i = 0; i < theInteraction.size(); i++ )
		{
			try
			{
				// print the parameter handle
				builder.append( "\tparamHandle=" );
				builder.append( theInteraction.getParameterHandle(i) );
				// print the parameter value
				builder.append( ", paramValue=" );
				builder.append(
				    EncodingHelpers.decodeString(theInteraction.getValue(i)) );
				builder.append( "\n" );
			}
			catch( ArrayIndexOutOfBounds aioob )
			{
				// won't happen
			}
		}

		log( builder.toString() );
	}

	public void removeObjectInstance( int theObject, byte[] userSuppliedTag )
	{
		log( "Object Removed: handle=" + theObject );
	}

	public void removeObjectInstance( int theObject,
	                                  byte[] userSuppliedTag,
	                                  LogicalTime theTime,
	                                  EventRetractionHandle retractionHandle )
	{
		log( "Object Removed: handle=" + theObject );
	}


	
}
